import sys
import unittest
import numpy
from cil.framework import DataContainer
from cil.framework import ImageData
from cil.framework import AcquisitionData
from cil.framework import ImageGeometry
from cil.framework import AcquisitionGeometry
from timeit import default_timer as timer


try:
    from ccpi.filters import regularisers
    from ccpi.filters.cpu_regularisers import TV_ENERGY
    has_regularisation_toolkit = True
except ImportError as ie:
    # raise ImportError(ie + "\n\n" + 
    #                   "This plugin requires the additional package ccpi-regularisation\n" +
    #                   "Please install it via conda as ccpi-regularisation from the ccpi channel\n"+
    #                   "Minimal version is 20.04")
    has_regularisation_toolkit = False
print ("has_regularisation_toolkit", has_regularisation_toolkit)

class TestPlugin(unittest.TestCase):
    def setUp(self):
        print ("test plugins")
        pass
    def tearDown(self):
        pass
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_FGP_TV(self):
        try:
            from cil.plugins.ccpi_regularisation.functions.regularisers import FGP_TV
            assert True
        except ModuleNotFoundError as ie:
            print (ie)
            assert False
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_ROF_TV(self):
        try:
            from cil.plugins.ccpi_regularisation.functions import ROF_TV
            assert True
        except ModuleNotFoundError as ie:
            assert False
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_TGV(self):
        try:
            from cil.plugins.ccpi_regularisation.functions import TGV
            assert True
        except ModuleNotFoundError as ie:
            assert False
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_LLT_ROF(self):
        try:
            from cil.plugins.ccpi_regularisation.functions import LLT_ROF
            assert True
        except ModuleNotFoundError as ie:
            assert False
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_FGP_dTV(self):
        try:
            from cil.plugins.ccpi_regularisation.functions import FGP_dTV
            assert True
        except ModuleNotFoundError as ie:
            assert False
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_SB_TV(self):
        try:
            from cil.plugins.ccpi_regularisation.functions import SB_TV
            assert True
        except ModuleNotFoundError as ie:
            assert False
    @unittest.skipUnless(has_regularisation_toolkit, "Skipping as CCPi Regularisation Toolkit is not installed")
    def test_import_TNV(self):
        try:
            from cil.plugins.ccpi_regularisation.functions import TNV
            assert True
        except ModuleNotFoundError as ie:
            assert False


# -*- coding: utf-8 -*-
#========================================================================
# Copyright 2018-2020 Science Technology Facilities Council
# Copyright 2018-2020 University of Manchester
#
# This work is part of the Core Imaging Library developed by Science Technology
# Facilities Council and University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#=========================================================================

import multiprocessing
NUM_THREADS = max(int(multiprocessing.cpu_count()/2),1)


# -*- coding: utf-8 -*-
#  CCP in Tomographic Imaging (CCPi) Core Imaging Library (CIL).

#   Copyright 2017 UKRI-STFC
#   Copyright 2017 University of Manchester

#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from cil.framework import Processor, AcquisitionData, ImageData
import warnings


class Resizer(Processor):

    def __init__(self,
                 roi = -1,
                 binning = 1):
        
        '''
        Constructor
        
        Input:
            
            roi             region-of-interest to crop. If roi = -1 (default), then no crop. 
                            Otherwise roi is given by a list with ndim elements, 
                            where each element is either -1 if no crop along this 
                            dimension or a tuple with beginning and end coodinates to crop to.
                            Example:
                                to crop 4D array along 2nd dimension:
                                roi = [-1, -1, (100, 900), -1]
                            
            binning         number of pixels to bin (combine) along each dimension.
                            If binning = 1, then projections in original resolution are loaded. 
                            Otherwise, binning is given by a list with ndim integers. 
                            Example:
                                to rebin 3D array along 1st direction:
                                binning = [1, 5, 1]
        '''

        kwargs = {'roi': roi,
                  'binning': binning}

        super(Resizer, self).__init__(**kwargs)
    
    def check_input(self, data):
        if not ((isinstance(data, ImageData)) or 
                (isinstance(data, AcquisitionData))):
            raise Exception('Processor supports only following data types:\n' +
                            ' - ImageData\n - AcquisitionData')
        elif (data.geometry == None):
            raise Exception('Geometry is not defined.')
        else:
            return True
    
    def process(self):

        data = self.get_input()
        ndim = len(data.dimension_labels)
        
        geometry_0 = data.geometry
        geometry = geometry_0.clone()
        
        if (self.roi == -1):
            roi_par = [-1] * ndim
        else:
            roi_par = list(self.roi)
            if (len(roi_par) != ndim):
                raise Exception('Number of dimensions and number of elements in roi parameter do not match')

        if (self.binning == 1):
            binning = [1] * ndim
        else:
            binning = list(self.binning)
            if (len(binning) != ndim):
                raise Exception('Number of dimensions and number of elements in binning parameter do not match')
                
        if (isinstance(data, ImageData)):
            if ((all(x == -1 for x in roi_par)) and (all(x == 1 for x in binning))):
                for key in data.dimension_labels:
                    if data.dimension_labels[key] == 'channel':
                        geometry.channels = geometry_0.channels
                        roi_par[key] = (0, geometry.channels)
                    elif data.dimension_labels[key] == 'horizontal_y':
                        geometry.voxel_size_y = geometry_0.voxel_size_y
                        geometry.voxel_num_y = geometry_0.voxel_num_y
                        roi_par[key] = (0, geometry.voxel_num_y)
                    elif data.dimension_labels[key] == 'vertical':
                        geometry.voxel_size_z = geometry_0.voxel_size_z
                        geometry.voxel_num_z = geometry_0.voxel_num_z
                        roi_par[key] = (0, geometry.voxel_num_z)
                    elif data.dimension_labels[key] == 'horizontal_x':
                        geometry.voxel_size_x = geometry_0.voxel_size_x
                        geometry.voxel_num_x = geometry_0.voxel_num_x
                        roi_par[key] = (0, geometry.voxel_num_x)
            else:
                for key in data.dimension_labels:
                    if data.dimension_labels[key] == 'channel':
                        if (roi_par[key] != -1):
                            geometry.set_channels ( (roi_par[key][1] - roi_par[key][0]) // binning[key] )
                            roi_par[key] = (roi_par[key][0], roi_par[key][0] + ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.set_channels ( geometry_0.channels // binning[key] )
                            roi_par[key] = (0, geometry.channels * binning[key])
                    elif data.dimension_labels[key] == 'horizontal_y':
                        if (roi_par[key] != -1):
                            geometry.voxel_num_y = (roi_par[key][1] - roi_par[key][0]) // binning[key]
                            geometry.voxel_size_y = geometry_0.voxel_size_y * binning[key]
                            roi_par[key] = (roi_par[key][0], roi_par[key][0] + ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.voxel_num_y = geometry_0.voxel_num_y // binning[key]
                            geometry.voxel_size_y = geometry_0.voxel_size_y * binning[key]
                            roi_par[key] = (0, geometry.voxel_num_y * binning[key])
                    elif data.dimension_labels[key] == 'vertical':
                        if (roi_par[key] != -1):
                            geometry.voxel_num_z = (roi_par[key][1] - roi_par[key][0]) // binning[key]
                            geometry.voxel_size_z = geometry_0.voxel_size_z * binning[key]
                            roi_par[key] = (roi_par[key][0], roi_par[key][0] + ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.voxel_num_z = geometry_0.voxel_num_z // binning[key]
                            geometry.voxel_size_z = geometry_0.voxel_size_z * binning[key]
                            roi_par[key] = (0, geometry.voxel_num_z * binning[key])
                    elif data.dimension_labels[key] == 'horizontal_x':
                        if (roi_par[key] != -1):
                            geometry.voxel_num_x = (roi_par[key][1] - roi_par[key][0]) // binning[key]
                            geometry.voxel_size_x = geometry_0.voxel_size_x * binning[key]
                            roi_par[key] = (roi_par[key][0], roi_par[key][0]+ ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.voxel_num_x = geometry_0.voxel_num_x // binning[key]
                            geometry.voxel_size_x = geometry_0.voxel_size_x * binning[key]
                            roi_par[key] = (0, geometry.voxel_num_x * binning[key])
            
        else: # AcquisitionData
            if ((all(x == -1 for x in roi_par)) and (all(x == 1 for x in binning))):
                for key in data.dimension_labels:
                    if data.dimension_labels[key] == 'channel':
                        geometry.channels = geometry_0.channels
                        roi_par[key] = (0, geometry.channels)
                    elif data.dimension_labels[key] == 'angle':
                        geometry.angles = geometry_0.angles
                        roi_par[key] = (0, len(geometry.angles))
                    elif data.dimension_labels[key] == 'vertical':
                        geometry.pixel_size_v = geometry_0.pixel_size_v
                        geometry.pixel_num_v = geometry_0.pixel_num_v
                        roi_par[key] = (0, geometry.pixel_num_v)
                    elif data.dimension_labels[key] == 'horizontal':
                        geometry.pixel_size_h = geometry_0.pixel_size_h
                        geometry.pixel_num_h = geometry_0.pixel_num_h
                        roi_par[key] = (0, geometry.pixel_num_h)
            else:
                for key in data.dimension_labels:
                    if data.dimension_labels[key] == 'channel':
                        if (roi_par[key] != -1):
                            geometry.channels = (roi_par[key][1] - roi_par[key][0]) // binning[key]
                            roi_par[key] = (roi_par[key][0], roi_par[key][0] + ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.channels = geometry_0.channels // binning[key]
                            roi_par[key] = (0, geometry.channels * binning[key])
                    elif data.dimension_labels[key] == 'angle':
                        if (roi_par[key] != -1):
                            geometry.set_angles( geometry_0.angles[roi_par[key][0]:roi_par[key][1]] )
                        else:
                            geometry.set_angles( geometry_0.angles )
                            roi_par[key] = (0, len(geometry.angles))
                        if (binning[key] != 1):
                            binning[key] = 1
                            warnings.warn('Rebinning in angular dimensions is not supported: \n binning[{}] is set to 1.'.format(key))
                    elif data.dimension_labels[key] == 'vertical':
                        if (roi_par[key] != -1):
                            geometry.pixel_num_v = (roi_par[key][1] - roi_par[key][0]) // binning[key]
                            geometry.pixel_size_v = geometry_0.pixel_size_v * binning[key]
                            roi_par[key] = (roi_par[key][0], roi_par[key][0] + ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.pixel_num_v = geometry_0.pixel_num_v // binning[key]
                            geometry.pixel_size_v = geometry_0.pixel_size_v * binning[key]
                            roi_par[key] = (0, geometry.pixel_num_v * binning[key])
                    elif data.dimension_labels[key] == 'horizontal':
                        if (roi_par[key] != -1):
                            geometry.pixel_num_h = (roi_par[key][1] - roi_par[key][0]) // binning[key]
                            geometry.pixel_size_h = geometry_0.pixel_size_h * binning[key]
                            roi_par[key] = (roi_par[key][0], roi_par[key][0] + ((roi_par[key][1] - roi_par[key][0]) // binning[key]) * binning[key])
                        else:
                            geometry.pixel_num_h = geometry_0.pixel_num_h // binning[key]
                            geometry.pixel_size_h = geometry_0.pixel_size_h * binning[key]
                            roi_par[key] = (0, geometry.pixel_num_h * binning[key])
                            
        if ndim == 2:
            n_pix_0 = (roi_par[0][1] - roi_par[0][0]) // binning[0]
            n_pix_1 = (roi_par[1][1] - roi_par[1][0]) // binning[1]
            shape = (n_pix_0, binning[0], 
                     n_pix_1, binning[1])
            data_resized = data.as_array()[roi_par[0][0]:(roi_par[0][0] + n_pix_0 * binning[0]), 
                                           roi_par[1][0]:(roi_par[1][0] + n_pix_1 * binning[1])].reshape(shape).mean(-1).mean(1)
        if ndim == 3:
            n_pix_0 = (roi_par[0][1] - roi_par[0][0]) // binning[0]
            n_pix_1 = (roi_par[1][1] - roi_par[1][0]) // binning[1]
            n_pix_2 = (roi_par[2][1] - roi_par[2][0]) // binning[2]
            shape = (n_pix_0, binning[0], 
                     n_pix_1, binning[1],
                     n_pix_2, binning[2])
            data_resized = data.as_array()[roi_par[0][0]:(roi_par[0][0] + n_pix_0 * binning[0]), 
                                           roi_par[1][0]:(roi_par[1][0] + n_pix_1 * binning[1]), 
                                           roi_par[2][0]:(roi_par[2][0] + n_pix_2 * binning[2])].reshape(shape).mean(-1).mean(1).mean(2)
        if ndim == 4:
            n_pix_0 = (roi_par[0][1] - roi_par[0][0]) // binning[0]
            n_pix_1 = (roi_par[1][1] - roi_par[1][0]) // binning[1]
            n_pix_2 = (roi_par[2][1] - roi_par[2][0]) // binning[2]
            n_pix_3 = (roi_par[3][1] - roi_par[3][0]) // binning[3]
            shape = (n_pix_0, binning[0], 
                     n_pix_1, binning[1],
                     n_pix_2, binning[2],
                     n_pix_3, binning[3])
            data_resized = data.as_array()[roi_par[0][0]:(roi_par[0][0] + n_pix_0 * binning[0]), 
                                           roi_par[1][0]:(roi_par[1][0] + n_pix_1 * binning[1]), 
                                           roi_par[2][0]:(roi_par[2][0] + n_pix_2 * binning[2]), 
                                           roi_par[3][0]:(roi_par[3][0] + n_pix_3 * binning[3])].reshape(shape).mean(-1).mean(1).mean(2).mean(3)

        out = type(data)(array = data_resized, 
                         deep_copy = False,
                         dimension_labels = data.dimension_labels,
                         geometry = geometry, 
                         suppress_warning=False)
        
        return out


'''
#usage exaample
ig = ImageGeometry(voxel_num_x = 200, 
                   voxel_num_y = 200, 
                   voxel_num_z = 200, 
                   voxel_size_x = 1, 
                   voxel_size_y = 1, 
                   voxel_size_z = 1, 
                   center_x = 0, 
                   center_y = 0, 
                   center_z = 0, 
                   channels = 200)

im = ImageData(array = numpy.zeros((200, 200, 200, 200)),
               geometry = ig,
               deep_copy = False, 
               dimension_labels = ['channel',\
                                   'vertical',\
                                   'horizontal_y',\
                                   'horizontal_x'])
            
            
resizer = Resizer(binning = [1, 1, 7, 1], roi = -1)
resizer.input = im
data_resized = resizer.process()
print(data_resized)
'''



# -*- coding: utf-8 -*-
#========================================================================
# Copyright 2019 Science Technology Facilities Council
# Copyright 2019 University of Manchester
#
# This work is part of the Core Imaging Library developed by Science Technology
# Facilities Council and University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0.txt
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#=========================================================================

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from cil.optimisation.functions import Function
from cil.framework import BlockDataContainer
from numbers import Number

class BlockFunction(Function):
    
    r""" BlockFunction represents a *separable sum* function :math:`F` defined as
    
    .. math:: F:X_{1}\times X_{2}\cdots\times X_{m} \rightarrow (-\infty, \infty]
    
    where :math:`F` is the separable sum of functions :math:`(f_{i})_{i=1}^{m}`,
    
    .. math:: F(x_{1}, x_{2}, \cdots, x_{m}) = \overset{m}{\underset{i=1}{\sum}}f_{i}(x_{i}), \mbox{ with } f_{i}: X_{i} \rightarrow (-\infty, \infty].
    
    A nice property (due to it's separability structure) is that the proximal operator 
    can be decomposed along the proximal operators of each function :math:`f_{i}`.
    
    .. math:: \mathrm{prox}_{\tau F}(x) = ( \mathrm{prox}_{\tau f_{i}}(x_{i}) )_{i=1}^{m}
    
    In addition, if :math:`\tau := (\tau_{1},\dots,\tau_{m})`, then 
    
    .. math:: \mathrm{prox}_{\tau F}(x) = ( \mathrm{prox}_{\tau_{i} f_{i}}(x_{i}) )_{i=1}^{m}
    
    """
    
    def __init__(self, *functions):
                
        super(BlockFunction, self).__init__()
        self.functions = functions      
        self.length = len(self.functions)
       
    @property        
    def L(self):
        # compute Lipschitz constant if possible
        tmp_L = 0
        for func in self.functions:
            if func.L is not None:  
                tmp_L += func.L                
            else:
                tmp_L = None 
                break 
        return tmp_L     
                                
    def __call__(self, x):
        
        r""" Returns the value of the BlockFunction :math:`F`
        
        .. math:: F(x) = \overset{m}{\underset{i=1}{\sum}}f_{i}(x_{i}), \mbox{ where } x = (x_{1}, x_{2}, \cdots, x_{m}), \quad i = 1,2,\dots,m
                
        Parameter:
            
            x : BlockDataContainer and must have as many rows as self.length

            returns ..math:: \sum(f_i(x_i))
            
        """
        
        if self.length != x.shape[0]:
            raise ValueError('BlockFunction and BlockDataContainer have incompatible size')
        t = 0                
        for i in range(x.shape[0]):
            t += self.functions[i](x.get_item(i))
        return t
    
    def convex_conjugate(self, x):
        
        r"""Returns the value of the convex conjugate of the BlockFunction at :math:`x^{*}`.       
        
            .. math:: F^{*}(x^{*}) = \overset{m}{\underset{i=1}{\sum}}f_{i}^{*}(x^{*}_{i})
            
            Parameter:
            
                x : BlockDataContainer and must have as many rows as self.length            
        
        """     
        
        if self.length != x.shape[0]:
            raise ValueError('BlockFunction and BlockDataContainer have incompatible size')
        t = 0              
        for i in range(x.shape[0]):
            t += self.functions[i].convex_conjugate(x.get_item(i))
        return t  
    
    def proximal(self, x, tau, out = None):
        
        r"""Proximal operator of the BlockFunction at x:
        
            .. math:: \mathrm{prox}_{\tau F}(x) =  (\mathrm{prox}_{\tau f_{i}}(x_{i}))_{i=1}^{m}
            
            Parameter:
            
                x : BlockDataContainer and must have as many rows as self.length            
        """
        
        if self.length != x.shape[0]:
            raise ValueError('BlockFunction and BlockDataContainer have incompatible size')        
        
        if out is None:
                
            out = [None]*self.length
            if isinstance(tau, Number):
                for i in range(self.length):
                    out[i] = self.functions[i].proximal(x.get_item(i), tau)
            else:
                for i in range(self.length):
                    out[i] = self.functions[i].proximal(x.get_item(i), tau.get_item(i))                        
            
            return BlockDataContainer(*out)  
        
        else:
            if isinstance(tau, Number):
                for i in range(self.length):
                    self.functions[i].proximal(x.get_item(i), tau, out[i])
            else:
                for i in range(self.length):
                    self.functions[i].proximal(x.get_item(i), tau.get_item(i), out[i])
            
            
    
    def gradient(self, x, out=None):
        
        r"""Returns the value of the gradient of the BlockFunction function at x.
        
        .. math:: F'(x) = [f_{1}'(x_{1}), ... , f_{m}'(x_{m})]        
        
        Parameter:
            
            x : BlockDataContainer and must have as many rows as self.length
                
        """        
        
        if self.length != x.shape[0]:
            raise ValueError('BlockFunction and BlockDataContainer have incompatible size')        
        
        out = [None]*self.length
        for i in range(self.length):
            out[i] = self.functions[i].gradient(x.get_item(i))
            
        return  BlockDataContainer(*out)     

    def proximal_conjugate(self, x, tau, out = None):
        
        r"""Proximal operator of the convex conjugate of BlockFunction at x:
        
            .. math:: \mathrm{prox}_{\tau F^{*}}(x) = (\mathrm{prox}_{\tau f^{*}_{i}}(x^{*}_{i}))_{i=1}^{m}
            
            Parameter:
            
                x : BlockDataContainer and must have as many rows as self.length            
        """

        if self.length != x.shape[0]:
            raise ValueError('BlockFunction and BlockDataContainer have incompatible size')

        if out is not None:
            if isinstance(tau, Number):
                for i in range(self.length):
                    self.functions[i].proximal_conjugate(x.get_item(i), tau, out=out.get_item(i))
            else:
                for i in range(self.length):
                    self.functions[i].proximal_conjugate(x.get_item(i), tau.get_item(i),out=out.get_item(i))
            
        else:
                
            out = [None]*self.length
            if isinstance(tau, Number):
                for i in range(self.length):
                    out[i] = self.functions[i].proximal_conjugate(x.get_item(i), tau)
            else:
                for i in range(self.length):
                    out[i] = self.functions[i].proximal_conjugate(x.get_item(i), tau.get_item(i))
            
            return BlockDataContainer(*out)
        
    def __getitem__(self, row):
        return self.functions[row]
                            
    
    
if __name__ == '__main__':
    
    M, N, K = 20,30,50
    
    from cil.optimisation.functions import L2NormSquared, MixedL21Norm, L1Norm
    from cil.framework import ImageGeometry, BlockGeometry
    from cil.optimisation.operators import GradientOperator, IdentityOperator, BlockOperator
    import numpy
    import numpy as np
    
    
    ig = ImageGeometry(M, N)
    BG = BlockGeometry(ig, ig)
    
    u = ig.allocate('random_int')
    B = BlockOperator( GradientOperator(ig), IdentityOperator(ig) )
    
    U = B.direct(u)
    b = ig.allocate('random_int')
    
    f1 =  10 * MixedL21Norm()
    f2 =  5 * L2NormSquared(b=b)    
    
    f = BlockFunction(f1, f2)
    print(f.L)
    
    f = BlockFunction(f2, f2)
    print(f.L)    
#    tau = 0.3
#    
#    print( " without out " )
#    res_no_out = f.proximal_conjugate( U, tau)
#    res_out = B.range_geometry().allocate()
#    f.proximal_conjugate( U, tau, out = res_out)
#    
#    numpy.testing.assert_array_almost_equal(res_no_out[0][0].as_array(), \
#                                            res_out[0][0].as_array(), decimal=4) 
#    
#    numpy.testing.assert_array_almost_equal(res_no_out[0][1].as_array(), \
#                                            res_out[0][1].as_array(), decimal=4) 
#
#    numpy.testing.assert_array_almost_equal(res_no_out[1].as_array(), \
#                                            res_out[1].as_array(), decimal=4)     
#    
#    
#
#    ig1 = ImageGeometry(M, N)               
#    ig2 = ImageGeometry(2*M, N)
#    ig3 = ImageGeometry(3*M, 4*N)
#    
#    bg = BlockGeometry(ig1,ig2,ig3)
#    
#    z = bg.allocate('random_int')
#    
#    f1 = L1Norm()
#    f2 = 5 * L2NormSquared()
#    
#    f = BlockFunction(f2, f2, f2 + 5, f1 - 4, f1)
#    
#    res = f.convex_conjugate(z)
#    
#    ##########################################################################
#    
#    
#    
#    
#    
#    
#    
##    zzz = B.range_geometry().allocate('random_int')
##    www = B.range_geometry().allocate()
##    www.fill(zzz)
#    
##    res[0].fill(z)
#    
#    
#    
#    
##    f.proximal_conjugate(z, sigma, out = res)
#    
##    print(z1[0][0].as_array())    
##    print(res[0][0].as_array())
#    
#    
#    
#    
##    U = BG.allocate('random_int')
##    RES = BG.allocate()
##    f = BlockFunction(f1, f2)
##    
##    z = f.proximal_conjugate(U, 0.2)
##    f.proximal_conjugate(U, 0.2, out = RES)
##    
##    print(z[0].as_array())
##    print(RES[0].as_array())
##    
##    print(z[1].as_array())
##    print(RES[1].as_array())    
#    
#    
#    
#    
#    
#
#
#    
